#include <QDebug>
#include <QMessageBox>
#include <QProcess>
#include <QDebug>

#include "maingui.h"
#include "ui_maingui.h"

/**
 * Constructeur
 *
 * @brief MainGUI::MainGUI
 * @param parent
 */
MainGUI::MainGUI(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::MainGUI)
{
    QString rpiModel;

    ui->setupUi(this);

    // Mettre en place les slots
    connect(this->ui->tabWidget, SIGNAL(currentChanged(int)), this, SLOT(onCurrentChanged(int)));
    // Signaux émis par l'onglet de définition du signal carré
    connect(this->ui->ledtSquareTh, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtSquareTl, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtSquareVh, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtSquareVl, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    // Signaux émis par l'onglet de définition du signal en triangle
    connect(this->ui->ledtTriangleTr, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtTriangleVh, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtTriangleVl, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    // Signaux émis par l'onglet de définition du signal en dents de scie
    connect(this->ui->ledtSawToothTr, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtSawToothVh, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtSawToothVl, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    // Signaux émis par l'onglet de définition du signal défini par l'utilisateur
    connect(this->ui->ledtUserTf, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtUserTh, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtUserTl, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtUserTr, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtUserVh, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    connect(this->ui->ledtUserVl, SIGNAL(textChanged(QString)), this, SLOT(onParamChanged(QString)));
    // Signal émis par le bouton de validation de la configuration pour la génération du signal DAC
    connect(this->ui->btnSetSignal, SIGNAL(clicked()), this, SLOT(onSetSignalClicked()));
    // Signal émis par le bouton de lancement/arrêt de la génération
    connect(this->ui->btnStartAndStop, SIGNAL(clicked(bool)), this, SLOT(onBtnStartAndStopClicked(bool)));
    // Signaux émis par les radio boutons de configuration de résolution du DAC
    connect(this->ui->rbtnDac12, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));
    connect(this->ui->rbtnDac14, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));
    connect(this->ui->rbtnDac16, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));
    // Signaux émis par les radio boutons de choix de canal du DAC
    connect(this->ui->rbtnDacA, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));
    connect(this->ui->rbtnDacB, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));
    connect(this->ui->rbtnDacC, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));
    connect(this->ui->rbtnDacD, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));
     // Signaux émis par les radio boutons de configuration du diviseur utilisé pour générer la Sclck
    connect(this->ui->rbtnSpiDiv8, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));
    connect(this->ui->rbtnSpiDiv16, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));
    connect(this->ui->rbtnSpiDiv32, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));
    connect(this->ui->rbtnSpiDiv64, SIGNAL(clicked()), this, SLOT(onSclckOrResolutionOrChannelChanged()));

    // Renseigner le texte dans la boite d'informations sur la SCLCK
    // . Déterminer le modèle de RPi
    rpiModel = this->getRpiModel();
    // . Autoriser l'ouverture des liens cliqués dans le navigateur
    //   par défaut
    this->ui->tbrwsrInfoSclck->setOpenExternalLinks(true);
    // . Peupler la boite avec le texte
    this->ui->tbrwsrInfoSclck->setText(
      "Il semble que le programme tourne sur une <b>Raspberry Pi "
      + rpiModel +"</b><br /><br />"
      + "Comme le mentionne la <a href=\"http://www.airspayce.com/mikem/bcm2835/group__constants.html#gaf2e0ca069b8caef24602a02e8a00884e\">documentation de la librairie bcm2835</a> "
      + "utilisée pour piloter le DAC à travers le bus SPI, "
      + "la fréquence du coeur est de " + (rpiModel == "3 model B" ? QString::number(400) : QString::number(250)) + "MHz sur ce modèle de Raspberry Pi<br /><br />"
      + "La fréquence du signal SCLK est obtenue par division de cette fréquence. Les 4 plus hautes fréquences possibles sont proposées ci-après."
      );

    // Modifier les valeurs affichées de la SCLCK selon le modèle de RPi
    if (rpiModel == "3 model B") {
        this->ui->rbtnSpiDiv8->setText("50MHz");
        this->ui->rbtnSpiDiv16->setText("25MHz");
        this->ui->rbtnSpiDiv32->setText("12.5MHz");
        this->ui->rbtnSpiDiv64->setText("6.250MHz");
    }

    // Sélectionner l'onglet du signal carré
    this->ui->tabWidget->setCurrentIndex(0);
    this->ui->btnSetSignal->setEnabled(true);

}

/**
 * Destructeur
 *
 * @brief MainGUI::~MainGUI
 */
MainGUI::~MainGUI()
{
    delete ui;
}


/**
 * Retourne le modèle de la raspberry pi sur laquelle tourne le programme
 *
 * @brief MainGUI::getRpiModel
 * @return
 */
QString MainGUI::getRpiModel() {
    QProcess cat;
    QProcess sed;
    QString revision;
    QList<QString> model3B = {
        "a02082", "a22082", "a32082"
    };
    QList<QString> model2B = {
        "a01040", "a01041", "a21041", "a22042"
    };
    QList<QString> model1A = {
        "7", "8", "9"
    };
    QList<QString> model1Aplus = {
        "12", "15", "900021"
    };
    QList<QString> model1B = {
        "2", "4", "5", "6", "000d", "000e", "000f", "Beta", "3"
    };
    QList<QString> model1Bplus = {
        "10", "13", "900032"
    };

    // Redirige la sortie du processus 'cat' sur l'entrée du
    // processus 'sed'
    cat.setStandardOutputProcess(&sed);

    // Lance le processus 'cat' pour récupérer la ligne contenant
    // le n° de révision de la RPi
    cat.start("cat /proc/cpuinfo");
    cat.waitForFinished(-1);

    // Lance le processus 'sed' avec une expression régulière étendue
    // pour extraire uniquement le n° de révision de la RPI
    // (Rem : la présence du nombre '1000' avant le n° de révision
    // indique que la RPi est overclockée)
    sed.start("sed -n -r -e\"s/^Revision.+\\b(1000)?(.+)$/\\2/p\"");
    sed.waitForFinished(-1);

    // Lit la sortie du processus 'sed'
    revision = QString(sed.readAllStandardOutput()).trimmed();
    qDebug() << "Revision RPi = " << revision << endl;


    if (model3B.lastIndexOf(revision) != -1) {
        return "3 model B";
    }

    if (model2B.lastIndexOf(revision) != -1) {
        return "2 model B";
    }

    if (model1A.lastIndexOf(revision) != -1) {
        return "1 model A";
    }

    if (model1Aplus.lastIndexOf(revision) != -1) {
        return "1 model A+";
    }

    if (model1B.lastIndexOf(revision) != -1) {
        return "1 model B";
    }

    if (model1Bplus.lastIndexOf(revision) != -1) {
        return "1 model B+";
    }

    return "unknown model";
}

/**
 * Slot appelé lors d'un changement d'onglet dans le widget de
 * sélection de signal
 * @brief MainGUI::onCurrentChanged
 * @param tabIdx
 */
void MainGUI::onCurrentChanged(int tabIdx) {
    tabIdx++; //uniquement pour ne pas avoir de warning du fait qu'on ne l'utilise pas
    this->ui->btnSetSignal->setEnabled(true);
    this->ui->btnStartAndStop->setEnabled(false);
}

/**
 * Slot appelé à chaque fois que la valeur d'un paramètre de signal change.
 * @brief MainGUI::onParamChanged
 * @param paramValue
 */
void MainGUI::onParamChanged(QString paramValue) {
    paramValue = "";
    this->ui->btnSetSignal->setEnabled(true);
    // Interdire la génération du signal tant que les paramètres
    // ne sont pas validés
    this->ui->btnStartAndStop->setEnabled(false);
}

/**
 * Slot appelé à chaque fois qu'un changement de résolution du DAC intervient
 */
void MainGUI::onSclckOrResolutionOrChannelChanged() {
    this->ui->btnSetSignal->setEnabled(true);
    // Interdire la génération du signal tant que les paramètres
    // ne sont pas validés
    this->ui->btnStartAndStop->setEnabled(false);
}

/**
 * Slot appelé sur validation des paramètres d'un signal à générer
 * @brief MainGUI::onSetSignalClicked
 */
void MainGUI::onSetSignalClicked() {
    int dacResolution;
    int dacChannel;
    int spiClockDiv;
    int tl, tr, th, tf;
    double vl, vh;
    bool tlOK, trOK, thOK, tfOK, vlOK, vhOK;

    switch(this->ui->tabWidget->currentIndex()) {
        case SQUARE :
            vl = this->checkAndConvertVoltage(
                        this->ui->ledtSquareVl->text()
                        , 0.0
                        , 5.0
                        , vlOK
                        );
            if (!vlOK) {
                break;
            }
            vh = this->checkAndConvertVoltage(
                        ui->ledtSquareVh->text()
                        , vl
                        , 5.0
                        , vhOK
                        );
            if(!vhOK) {
                break;
            }
            tr = tf = 0;
            trOK = tfOK = true;
            tl = this->checkAndConvertTime(
                        this->ui->ledtSquareTl->text()
                        , tlOK
                        );
            if(!tlOK) {
                break;
            }
            th = this->checkAndConvertTime(
                        this->ui->ledtSquareTh->text()
                        ,thOK
                        );

            break;

        case TRIANGLE :
            vl = this->checkAndConvertVoltage(
                        this->ui->ledtTriangleVl->text()
                        , 0.0
                        , 5.5
                        , vlOK
                        );
            if(!vlOK) {
                break;
            }
            vh = this->checkAndConvertVoltage(
                        ui->ledtTriangleVh->text()
                        , vl
                        , 5.0
                        , vhOK
                        );
            if(!vhOK) {
                break;
            }
            tr = this->checkAndConvertTime(
                        this->ui->ledtTriangleTr->text()
                        , trOK
                        );
            if(!trOK) {
                break;
            }
            tf = tr;
            tfOK = trOK;
            tl = th = 0;
            tlOK = thOK = true;
            break;

        case SAWTOOTH :
            vl = this->checkAndConvertVoltage(
                        this->ui->ledtSawToothVl->text()
                        , 0.0
                        , 5.0
                        , vlOK
                        );
            if (!vlOK) {
                break;
            }
            vh = this->checkAndConvertVoltage(
                        ui->ledtSawToothVh->text()
                        , vl
                        , 5.0
                        , vhOK
                        );
            tr = this->checkAndConvertTime(
                        this->ui->ledtSawToothTr->text()
                        , trOK
                        );
            tf = 0;
            tfOK = true;
            tl = th = 0;
            tlOK = thOK = true;;
            break;

        case USER_DEFINED :
            vl = this->checkAndConvertVoltage(
                        this->ui->ledtUserVl->text()
                        , 0.0
                        , 5.0
                        , vlOK
                        );
            if (!vlOK) {
                break;
            }
            vh = this->checkAndConvertVoltage(
                        ui->ledtUserVh->text()
                        , vl
                        , 5.0
                        , vhOK
                        );
            if(!vhOK) {
                break;
            }
            tl = this->checkAndConvertTime(
                        this->ui->ledtUserTl->text()
                        , tlOK
                        );
            if(!tlOK) {
                break;
            }
            tr = this->checkAndConvertTime(
                        this->ui->ledtUserTr->text()
                        , trOK
                        );
            if(!trOK) {
                break;
            }
            th = this->checkAndConvertTime(
                        this->ui->ledtUserTh->text()
                        , thOK
                        );
            if(!thOK) {
                break;
            }
            tf = this->checkAndConvertTime(
                        this->ui->ledtUserTf->text()
                        , tfOK
                        );
            break;
    }

    if( tlOK && trOK && thOK && tfOK && vlOK && vhOK ) {
        // Fixe la résolution du DAC
        // (Ceci est à faire avant d'appeler pg.computeSignal())
        dacResolution = this->ui->rbtnDac12->isChecked() ? 12 : -1;
        dacResolution = this->ui->rbtnDac14->isChecked() ? 14 : dacResolution;
        dacResolution = this->ui->rbtnDac16->isChecked() ? 16 : dacResolution;
        this->pg.setDacResolution(dacResolution);

        // Fixe le canal du DAC sur lequel le signal doit être généré
        dacChannel = this->ui->rbtnDacA->isChecked() ? 0 : -1;
        dacChannel = this->ui->rbtnDacB->isChecked() ? 1 : dacChannel;
        dacChannel = this->ui->rbtnDacC->isChecked() ? 2 : dacChannel;
        dacChannel = this->ui->rbtnDacD->isChecked() ? 3 : dacChannel;
        this->pg.setDacChannel(dacChannel);

		  // Calculer les points du signal à générer
        pg.computeSignal(tl, tr, th, tf, vl, vh);

		  // Fixer la fréquence du signal Sclck
        spiClockDiv = this->ui->rbtnSpiDiv8->isChecked() ? 8 : -1;
        spiClockDiv = this->ui->rbtnSpiDiv16->isChecked() ? 16 : spiClockDiv;
        spiClockDiv = this->ui->rbtnSpiDiv32->isChecked() ? 32 : spiClockDiv;
        spiClockDiv = this->ui->rbtnSpiDiv64->isChecked() ? 64 : spiClockDiv;
        this->pg.setSclck(spiClockDiv);

  		  // Autoriser la génération 
        this->ui->btnSetSignal->setEnabled(false);
        this->ui->btnStartAndStop->setEnabled(true);
        //qDebug() << "vl :"  << vl << " / vh : "  << vh << "tl : "  <<tl << " / tr : " << tr << " / th : " << th << " /tf : "<< tf << endl;
    }

}

/**
 * Slot qui permet de lancer/stopper la génération du signal DAC
 * @brief MainGUI::onBtnStartAndStopClicked
 */
void MainGUI::onBtnStartAndStopClicked(bool isChecked) {
    if(isChecked) {
        this->ui->grpbxSelectSclck->setEnabled(false);
        this->ui->grpbxSelectResolution->setEnabled(false);
        this->ui->grpbxSelectDacChannel->setEnabled(false);
        this->ui->grpbxSelectWave->setEnabled(false);
        this->pg.start();
        this->ui->btnStartAndStop->setText("Stopper");
      } else {
        this->ui->grpbxSelectSclck->setEnabled(true);
        this->ui->grpbxSelectResolution->setEnabled(true);
        this->ui->grpbxSelectDacChannel->setEnabled(true);
        this->ui->grpbxSelectWave->setEnabled(true);
        this->pg.stop();
        this->ui->btnStartAndStop->setText("Générer");
    }
}

/**
 * Contrôle la saisie pour une tension.
 * @brief MainGUI::checkAndConvertVoltage
 * @param str
 * @param min
 * @param max
 * @param ok
 * @return
 */
double MainGUI::checkAndConvertVoltage(QString str, double min, double max, bool &ok) {
    double val = str.toDouble(&ok);
    QString msg;

     msg  = "La valeur '" + str + "' dans le champ de saisie n'est pas valide !\n";
     msg += "Elle doit être de type décimal et comprise entre " + QString::number(min) + " et " + QString::number(max);

    if ((ok == false) || (val < min) || (val > max)) {
        QMessageBox::critical(this, "Valeur non valide", msg);
        val = -1;
    }

    return val;
}

/**
 * Contrôle la saisie pour une durée
 * @brief MainGUI::checkAndConvertTime
 * @param str
 * @param ok
 * @return
 */
int MainGUI::checkAndConvertTime(QString str, bool &ok) {
    double val = str.toInt(&ok);
    QString msg;

     msg  = "La valeur '" + str + "' dans le champ de saisie n'est pas valide !\n";
     msg += "Elle doit être de type entier positif";

    if (!ok || val < 0) {
        QMessageBox::critical(this, "Valeur non valide", msg);
        val = -1;
    }

    return val;
}
