#ifndef PULSEGENERATOR_H
#define PULSEGENERATOR_H

#include <cstdint>
#include <thread>
#include <atomic>
#include <chrono>
#include <vector>

#include <bcm2835.h>

class PulseGenerator
{
public:
    PulseGenerator();
    ~PulseGenerator();
    typedef enum {RISING, FALLING} edge_t;
    void start();
    int stop();
    void setDacResolution(int resolution);
    void setDacChannel(int channel);
    void setSclck(int div);
    void computeSignal(int tl, int tr, int th, int tf, double vl, double vh);
    void computeEdge(u_int16_t vi, u_int16_t vf, int t, edge_t edgeType);
private:
    const std::uint8_t CS_LD = BCM2835_SPI_CS0;

	 std::atomic<bool> quit;
    std::thread t;
    void generatorThread();

    double vRef;
    u_int16_t fullScale;
    int resolution;
    int channel;
    int tLow;
    int tRising;
    int tHigh;
    int tFalling;
    u_int16_t vLow;
    u_int16_t vHigh;
    std::vector<u_int16_t> risindEdge;
    std::vector<u_int16_t> fallingEdge;
    std::vector<u_int16_t> risindEdgeBresenham;

    void rasterizeWithLinearEquation(u_int16_t xi, u_int16_t yi, u_int16_t xf, u_int16_t yf, std::vector<u_int16_t> * dots);
    void rasterizeWithBresenhamAlgorithm(u_int16_t xi, u_int16_t yi, u_int16_t xf, u_int16_t yf, std::vector<u_int16_t> & dots);
};

#endif // PULSEGENERATOR_H
